"""
This code has been developed by Carl Stahoviak and then modified by Samuel Lovett

Original code available at https://github.com/cstahoviak/goggles
"""

from __future__ import division
import numpy as np
from functools import reduce



## filtering of 2D and 3D radar data
def AIRE_filtering(data_AIRE, thresholds):
    ## unpack data (into row vectors)
    radar_azimuth = data_AIRE[:, 0]  # [rad]
    radar_intensity = data_AIRE[:, 1]  # [dB]
    radar_range = data_AIRE[:, 2]  # [m]
    radar_elevation = data_AIRE[:, 3]  # [rad]
    radar_doppler = data_AIRE[:, 3]  # [m/s]

    azimuth_thres = thresholds[0]  # [deg]
    intensity_thres = thresholds[1]  # [dB]
    range_thres = thresholds[2]  # [m]
    range_thres_max = thresholds[3]  # [m]
    elevation_thres = thresholds[4]  # [deg]
    doppler_thres = thresholds[5]  # [m/s]

    ## Indexing in Python example
    ## print("Values bigger than 10 =", x[x>10])
    ## print("Their indices are ", np.nonzero(x > 10))
    idx_azimuth = np.nonzero(np.abs(np.rad2deg(radar_azimuth)) < azimuth_thres)
    idx_intensity = np.nonzero(radar_intensity > intensity_thres)
    idx_range = np.nonzero(radar_range > range_thres)
    idx_range_max = np.nonzero(radar_range < range_thres_max)
    idx_doppler = np.nonzero(np.abs(radar_doppler < doppler_thres))

    ## 3D radar data
    idx_elevation = np.nonzero(np.abs(np.rad2deg(radar_elevation)) < elevation_thres)
    idx_AIRE = reduce(np.intersect1d, (idx_azimuth, idx_intensity, idx_range,
                                       idx_range_max, idx_elevation, idx_doppler))
    Ntargets_valid = idx_AIRE.shape[0]

    return idx_AIRE, Ntargets_valid


"""
Author:         Samuel Lovett
Date Created:   Apr 26, 2024
Last Edited:    Dec 06, 2024

Description:
RANdom SAmple Consensus program described in Sec. III of [1]

[1] Temporally Constrained Instanteneous Ego-Motion Estimation using 3D Doppler Radar
"""
from tempEgo3D.data_unpacker import data_unpacker_tensor_object_3D
from tempEgo3D.least_squares_3D import least_squares_velocity_analysis_3D
import numpy as np

rng = np.random.default_rng()

class RANSAC:
    """
    Standard ransac class which implements ransac method as shown in Sec. III A. of [1]
    """

    def __init__(self, n=10, k=100, epsilon=0.05, rho=10, loss=None, metric=None):
        """
        Initialization of the standard velocity estimation method (KB) as described at the beginning of this code.
        :param n: Minimum number of samples to estimate parameters
        :param k: Maximum iterations allowed
        :param epsilon: Threshold value to determine if points are fit well
        :param rho: Number of close samples required to assert model fits well
        :param loss: function of `y_true` and `y_pred` that returns a vector scoring how well y_pred matches y_true at each point
        :param metric: function of `y_true` and `y_pred` and returns a float scoring how well y_pred matches y_true overall
        """
        self.n = n
        self.k = k
        self.epsilon = epsilon
        self.rho = rho
        self.loss = loss
        self.metric = metric

    def estimate_velocity(self, buffer):
        """
        Filter the inlier point from the outlier points using RANSAC and return the body velocity. \n
        return vel: list containing the x, y, and z body velocity [velocity in x, velocity in y, velocity in z]
        :param buffer: multidimensional list of measurement data [theta, gamma, doppler] (3, number of samples)
        :return: vel
        """
        measured_azim_val, measured_elev_val, measured_doppler_val, _, _ = buffer
        best_error = np.inf
        index = np.arange(len(measured_azim_val))
        maybe_model = SinusoidalRegressor()
        refined_model = SinusoidalRegressor()
        model_using_all_found_inliers_model = SinusoidalRegressor()

        n = self.n
        epsilon = self.epsilon
        rho = self.rho
        for _ in range(self.k):

            # create the list initial guess inliers by taking the n randomly selected indices
            # and copying the azimuth elevation and doppler values at those locations
            initial_guess_idx = rng.choice(index, n, replace=False, axis=0, shuffle=False)
            guess_inliers_azim = measured_azim_val[initial_guess_idx[:]]
            guess_inliers_elev = measured_elev_val[initial_guess_idx[:]]
            guess_inliers_doppler = measured_doppler_val[initial_guess_idx[:]]

            # make a model with our initial guess of inliers
            maybe_model.fit(guess_inliers_azim, guess_inliers_elev, guess_inliers_doppler)

            # using the created model predict the doppler values at the location of our azimuth values
            model_predicted_doppler_at_given_angles = maybe_model.predict(measured_azim_val, measured_elev_val)

            # create a boolean array that is the same length as our data arrays
            # compare the squared error distance between the predicted value and the measured value.
            # If it is greater than the threshold set that index value to true and if it is lower set the value to false
            thresholded = (self.loss(measured_doppler_val, model_predicted_doppler_at_given_angles) < epsilon)

            # copy the values from the index corresponding to the location of the true values within threshold
            inlier_ids = index[:][np.where(thresholded)[0]]
            number_of_inliers = inlier_ids.size

            # check if the number of inliers from this guess is higher than required to assert model fits well
            if number_of_inliers > rho:
                # fit a model with all the inliers found from the initial guess
                refined_model.fit(measured_azim_val[inlier_ids],
                                  measured_elev_val[inlier_ids],
                                  measured_doppler_val[inlier_ids])

                # for the subset of inliers, calculate the mean squared error between the measured doppler and
                # the predicted doppler from the better model (model using all inlier points)
                this_error = self.metric(measured_doppler_val[inlier_ids],
                                         refined_model.predict(measured_azim_val[inlier_ids],
                                                               measured_elev_val[inlier_ids]))

                # check if the error is less than the best error found so far
                if this_error <= best_error:
                    # update values since this model is better than the last
                    best_error = this_error
                    model_using_all_found_inliers_model = refined_model

        vel = [model_using_all_found_inliers_model.velocity_x,
               model_using_all_found_inliers_model.velocity_y,
               model_using_all_found_inliers_model.velocity_z]

        return vel


class TEMPSAC(RANSAC):
    """
    TEMPSAC class which implements TEMPSAC in Sec. III C. of [1]
    """

    def __init__(self, n=10, k=100, epsilon=0.05, rho=10, loss=None, metric=None):
        super().__init__(n, k, epsilon, rho, loss, metric)

    def get_initial_guess(self, buffer, probability):
        """
        Uses the probabilities to generate the likelihood of selecting a sample from a specific measurement and then uniformly selects samples from the selected measurement.
        The index of the selected samples are then returned for model fitting.
        :param buffer: samples being selected on
        :param probability: weights used to select the samples
        :return: n_random_points_full_idx
        """
        buffer_length = len(buffer)
        buffer_index = np.arange(buffer_length)
        try:
            n_random_buffer_idx = rng.choice(buffer_index, self.n, True, probability, 0, False)
        except Exception as e:
            print(e)
            exit()

        n_random_points_full_idx = []
        buffer_item_bins = []

        # find the length of each element within the buffer
        for m in range(buffer_length):
            number_of_data_points = len(buffer[m][0])
            buffer_item_bins.append(number_of_data_points)

        # Return the number of times each measurement is selected
        unique_buffer_idx, times_selected = np.unique(n_random_buffer_idx, return_counts=True)

        # Select the random samples from within each selected measurement in buffer
        for i in range(len(unique_buffer_idx)):
            num_points_to_select = times_selected[i]
            selected_buffer = unique_buffer_idx[i]
            measurement_idx = np.arange(buffer_item_bins[selected_buffer])
            random_points = rng.choice(measurement_idx, num_points_to_select, replace=False, shuffle=False)

            # Convert the within measurement index into the full index
            if unique_buffer_idx[i] == 0:
                n_random_points_full_idx.extend(random_points)
            else:
                for point_selected in random_points:
                    full_index = sum(buffer_item_bins[:selected_buffer]) + point_selected
                    n_random_points_full_idx.append(full_index)
        return n_random_points_full_idx

    def estimate_velocity(self, buffer):
        """
        Filter the inlier point from the outlier points using TEMPSAC and return the body velocity. \n
        return vel: list containing the x, y, and z body velocity [velocity in x, velocity in y, velocity in z]
        :param buffer: multidimensional list of measurement data [theta, gamma, doppler] (3, number of samples)
        :return: vel
        """
        buffer, probability = buffer
        measured_azim_val, measured_elev_val, measured_doppler_val, = data_unpacker_tensor_object_3D(buffer)
        best_error = np.inf
        epsilon = self.epsilon
        rho = self.rho

        index = np.arange(len(measured_azim_val))
        maybe_model = SinusoidalRegressor()
        better_model = SinusoidalRegressor()
        model_using_all_found_inliers_model = SinusoidalRegressor()

        for _ in range(self.k):

            # if you get an error about this line of code check your n value
            initial_guess_idx = self.get_initial_guess(buffer, probability)

            # create the list initial guess inliers by taking the n randomly selected indices
            # and copying the azimuth and doppler values at those locations
            guess_inliers_azim = measured_azim_val[initial_guess_idx[:]]
            guess_inliers_elev = measured_elev_val[initial_guess_idx[:]]
            guess_inliers_doppler = measured_doppler_val[initial_guess_idx[:]]

            # make a model with our initial guess of inliers
            maybe_model.fit(guess_inliers_azim, guess_inliers_elev, guess_inliers_doppler)

            # using the created model predict the doppler values at the location of our azimuth values
            model_predicted_doppler_at_given_angles = maybe_model.predict(measured_azim_val, measured_elev_val)

            # create a boolean array that is the same length as our data arrays
            # compare the squared error distance between the predicted value and the measured value.
            # If it is greater than the threshold set that index value to true and if it is lower set the value to false
            thresholded = (self.loss(measured_doppler_val, model_predicted_doppler_at_given_angles) < epsilon)

            # copy the values from the index corresponding to the location of the true values within threshold
            inlier_ids = index[:][np.where(thresholded)[0]]
            number_of_inliers = inlier_ids.size

            # check if the number of inliers from this guess is higher than required to assert model fits well
            if number_of_inliers > rho:
                # fit a model with all the inliers found from the initial guess
                better_model.fit(measured_azim_val[inlier_ids],
                                 measured_elev_val[inlier_ids],
                                 measured_doppler_val[inlier_ids])

                # for the subset of inliers, calculate the mean squared error between the measured doppler and
                # the predicted doppler from the better model (model using all inlier points)
                this_error = self.metric(measured_doppler_val[inlier_ids],
                                         better_model.predict(measured_azim_val[inlier_ids],
                                                              measured_elev_val[inlier_ids]))

                # check if the error is less than the best error found so far
                if this_error <= best_error:
                    # update values since this model is better than the last
                    best_error = this_error
                    model_using_all_found_inliers_model = better_model

        vel = [model_using_all_found_inliers_model.velocity_x,
               model_using_all_found_inliers_model.velocity_y,
               model_using_all_found_inliers_model.velocity_z]

        return vel


class RANSAC_TWLSQ(RANSAC):
    """
    TWLSQ ransac class which implements ransac identically to the standard method but passes the weights to the regressor class as shown in Sec. III D. of [1]
    """

    def __init__(self, n=10, k=100, epsilon=0.05, rho=10, loss=None, metric=None):
        super().__init__(n, k, epsilon, rho, loss, metric)

    def estimate_velocity(self, buffer):
        """
        Filter the inlier point from the outlier points using TWLSQ return the body velocity. \n
        return vel: list containing the x, y, and z body velocity [velocity in x, velocity in y, velocity in z]
        :param buffer: multidimensional list of measurement data [theta, gamma, doppler] (3, number of samples)
        :return: vel
        """
        measured_azim_val, measured_elev_val, measured_doppler_val, full_weight_mask = buffer
        best_error = np.inf
        index = np.arange(len(measured_azim_val))
        maybe_model = SinusoidalRegressor()
        better_model = SinusoidalRegressor()
        model_using_all_found_inliers_model = SinusoidalRegressor()

        n = self.n
        epsilon = self.epsilon
        rho = self.rho
        for _ in range(self.k):

            # create the list initial guess inliers by taking the n randomly selected indices
            # and copying the azimuth and doppler values at those locations
            initial_guess_idx = rng.choice(index, n, replace=False, axis=0, shuffle=False)
            guess_inliers_azim = measured_azim_val[initial_guess_idx[:]]
            guess_inliers_elev = measured_elev_val[initial_guess_idx[:]]
            guess_inliers_doppler = measured_doppler_val[initial_guess_idx[:]]

            # make a model with our initial guess of inliers
            weight_mask = full_weight_mask[initial_guess_idx[:]]
            maybe_model.fit(guess_inliers_azim, guess_inliers_elev, guess_inliers_doppler, weights=weight_mask)

            # using the created model predict the doppler values at the location of our azimuth values
            model_predicted_doppler_at_given_angles = maybe_model.predict(measured_azim_val, measured_elev_val)

            # create a boolean array that is the same length as our data arrays
            # compare the squared error distance between the predicted value and the measured value.
            # If it is greater than the threshold set that index value to true and if it is lower set the value to false
            thresholded = (self.loss(measured_doppler_val, model_predicted_doppler_at_given_angles) < epsilon)

            # copy the values from the index corresponding to the location of the true values within threshold
            inlier_ids = index[:][np.where(thresholded)[0]]
            number_of_inliers = inlier_ids.size

            # check if the number of inliers from this guess is higher than required to assert model fits well
            if number_of_inliers > rho:
                # fit a model with all the inliers found from the initial guess
                weight_mask = full_weight_mask[inlier_ids[:]]
                better_model.fit(measured_azim_val[inlier_ids],
                                 measured_elev_val[inlier_ids],
                                 measured_doppler_val[inlier_ids],
                                 weights=weight_mask)

                # for the subset of inliers, calculate the mean squared error between the measured doppler and
                # the predicted doppler from the better model (model using all inlier points)
                this_error = self.metric(measured_doppler_val[inlier_ids],
                                         better_model.predict(measured_azim_val[inlier_ids],
                                                              measured_elev_val[inlier_ids]))

                # check if the error is less than the best error found so far
                if this_error <= best_error:
                    # update values since this model is better than the last
                    best_error = this_error
                    model_using_all_found_inliers_model = better_model


        vel = [model_using_all_found_inliers_model.velocity_x,
               model_using_all_found_inliers_model.velocity_y,
               model_using_all_found_inliers_model.velocity_z]

        return vel



class SinusoidalRegressor:
    """
    Velocity model class which implements (13) from [1].
    """

    def __init__(self):
        self.velocity_x = 0
        self.velocity_y = 0
        self.velocity_z = 0
        self.weights = None

    def fit(self, azimuth, elev, doppler, weights=None):
        """
        Fit the model using LSQ or TWLSQ.
        \n param[in] azimuth: array of azimuth values.
        \n param[in] elev: array of elevation values.
        \n param[in] doppler: array of doppler values.
        \n param[in] weights: Optional array of weights corresponding to each sample.
        \n return self: return the parameterized velocity model object
        :param azimuth:
        :param elev:
        :param doppler:
        :param weights:
        :return: velocity model object
        """
        vx, vy, vz = least_squares_velocity_analysis_3D(azimuth, elev, doppler, weights)
        # Make velocity negative since it is moving in opposite direction of sensor
        # Sentence after equations (1) in "Instantaneous ego-motion estimation using Doppler radar" D. Kellner et al
        self.velocity_x = -1 * vx
        self.velocity_y = -1 * vy
        self.velocity_z = -1 * vz
        return self

    def predict(self, theta: np.ndarray, elev: np.ndarray):
        """
        Predict the doppler velocity from the given azimuth values using (13) from [1]
        :param theta: azimuth values to estimate the doppler velocity for
        :param elev: elevation values to estimate the doppler velocity for
        :return: doppler velocity estimates
        """

        predicted_velocity = -self.velocity_x * np.cos(elev) * np.cos(theta) \
                             + -self.velocity_y * np.cos(elev) * np.sin(theta) \
                             + -self.velocity_z * np.sin(elev)

        return predicted_velocity



"""
This code has been developed by
Kramer A, Harlow K, Williams C, Heckman C. ColoRadar: The direct 3D millimeter wave radar dataset.
The International Journal of Robotics Research. 2022;41(4):351-360. doi:10.1177/02783649211068535

Available at https://arpg.github.io/coloradar/
"""

import numpy as np
import struct
import os


# reads raw adc data from bin file 
# param[in] index: index of the requested frame
# param[in] base_dir: base directory of the sequence
# param[in] params: dict object containing radar config parameters
# return    numpy array containing complex adc samples with dimensions 
#            num_tx * num_rx * chirps_per_frame * adc_samples_per_chirp
def get_adc_frame(index, seq_dir, params):
    if params['sensor_type'] == 'cascade':
        filename = seq_dir + '/cascade/adc_samples/data/frame_' + str(index) + '.bin'
    else:
        filename = seq_dir + '/single_chip/adc_samples/data/frame_' + str(index) + '.bin'

    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return

    with open(filename, mode='rb') as file:
        frame_bytes = file.read()

    frame_vals = struct.unpack(str(len(frame_bytes) // 2) + 'h', frame_bytes)
    frame_vals = np.array(frame_vals, dtype=float)
    frame_vals = frame_vals[:-1:2] + 1j * frame_vals[1::2]

    frame = frame_vals.reshape(params['num_tx'], params['num_rx'], params['num_chirps_per_frame'],
                               params['num_adc_samples_per_chirp'])

    return frame


# reads post-angle-fft heatmap from bin file
# param[in] index: index of the requested heatmap
# param[in] seq_dir: base directory of the sequence
# param[in] params: dict object containing radar config parameters
# return    numpy array containing real-valued intensities for the post-angle-fft heatmap
#           with dimensions num_elevation_bins * num_azimuth_bins * num_range_bins * 2
#           where each elevation-azimuth-range bin has 2 values: the peak intensity and
#           the peak location for the doppler spectrum for that bin
def get_heatmap(index, seq_dir, params):
    if params['sensor_type'] == 'cascade':
        filename = seq_dir + '/cascade/heatmaps/data/heatmap_' + str(index) + '.bin'
    else:
        filename = seq_dir + '/single_chip/heatmaps/data/heatmap_' + str(index) + '.bin'

    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return None

    with open(filename, mode='rb') as file:
        frame_bytes = file.read()

    frame_vals = struct.unpack(str(len(frame_bytes) // 4) + 'f', frame_bytes)
    frame_vals = np.array(frame_vals)

    frame = frame_vals.reshape((params['num_elevation_bins'],
                                params['num_azimuth_bins'],
                                params['num_range_bins'],
                                2))  # 2 vals for each bin (doppler peak intensity and peak location)

    return frame


# reads pointcloud from bin file
# param[in] index: index of the requested pointcloud
# param[in] seq_dir: base directory of the sequence
# param[in] params: dict object containing sensor config parameters
# return    a numpy array containing one point per row, 
#           lidar point values are [x,y,z,intensity]
#           radar point values are [x,y,z,intensity,doppler]
def get_pointcloud(index, seq_dir, params):
    if params['sensor_type'] == 'lidar':
        filename = seq_dir + 'lidar/pointclouds/lidar_pointcloud_' + str(index) + '.bin'
    else:
        filename = seq_dir + 'single_chip/pointclouds/data/radar_pointcloud_' + str(index) + '.bin'

    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return None

    with open(filename, mode='rb') as file:
        cloud_bytes = file.read()

    cloud_vals = struct.unpack(str(len(cloud_bytes) // 4) + 'f', cloud_bytes)
    cloud_vals = np.array(cloud_vals)

    if params['sensor_type'] == 'lidar':
        cloud = cloud_vals.reshape((-1, 4))
    else:
        cloud = cloud_vals.reshape((-1, 5))

    return cloud


# reads all timestamps from file
# param[in] seq_dir: base directory of the sequence
# param[in] params: dict object containing sensor config parameters
# return    list of timestamps in seconds as floating point values
def get_timestamps(seq_dir, params):
    if params['sensor_type'] == 'lidar':
        filename = seq_dir + '/lidar/timestamps.txt'
    elif params['sensor_type'] == 'imu':
        filename = seq_dir + '/imu/timestamps.txt'
    elif params['sensor_type'] == 'groundtruth':
        filename = seq_dir + '/groundtruth/timestamps.txt'
    else:
        if params['sensor_type'] == 'cascade':
            filename = seq_dir + '/cascade'
        else:
            filename = seq_dir + '/single_chip'
        if params['data_type'] == 'adc_samples':
            filename += '/adc_samples/timestamps.txt'
        elif params['data_type'] == 'heatmap':
            filename += '/heatmaps/timestamps.txt'
        else:
            filename += '/pointclouds/timestamps.txt'

    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return

    with open(filename, mode='r') as file:
        lines = file.readlines()

    stamps = [float(line) for line in lines]

    return stamps


# reads all imu measurements from file
# param[in] seq_dir: base directory of the sequence
# return    list of imu measurements where each imu measurement is a dict with two entries:
#            {'accel': [x,y,z], 'gyro': [x,y,z]}
def get_imu(seq_dir):
    filename = seq_dir + '/imu/imu_data.txt'

    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return

    with open(filename, mode='r') as file:
        lines = file.readlines()

    accel_data = []
    gyro_data = []
    # imu_data = []
    for line in lines:
        vals = [float(s) for s in line.split()]
        # imu_data.append({'accel': vals[:3], 'gyro': vals[3:]})
        accel_data.append(vals[:3])
        gyro_data.append(vals[3:])

    return accel_data, gyro_data


# reads all groundtruth poses from file
# param[in] seq_dir: base directory of the sequence
# return    list of groundtruth poses where each pose is a dict with two entries:
#           {'position': [x,y,z], 'orientation': [x,y,z,w]}
def get_groundtruth(seq_dir):

  filename = seq_dir + '/groundtruth/groundtruth_poses.txt'

  if not os.path.exists(filename):
    print('File ' + filename + ' not found')
    return

  with open(filename, mode='r') as file:
    lines = file.readlines()

  gt_data = []
  for line in lines:
    vals = [float(s) for s in line.split()]
    gt_data.append({'position': np.array(vals[:3]),
                    'orientation': np.array(vals[3:])})

  return gt_data


# reads all vicon poses from file
# param[in] seq_dir: base directory of the sequence
# return    list of vicon poses where each pose is a dict with two entries:
#           {'position': [x,y,z], 'orientation': [x,y,z,w]}
def get_vicon(seq_dir):
    filename = seq_dir + '/vicon/vicon_poses.txt'

    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return

    with open(filename, mode='r') as file:
        lines = file.readlines()

    vicon_data = []
    for line in lines:
        vals = [float(s) for s in line.split()]
        vicon_data.append({'position': np.array(vals[:3]),
                           'orientation': np.array(vals[3:])})

    return vicon_data


# reads extrinsic transform data from a file
# param[in] filename: filename for the extrinsic transform file
# return t: the translation [x,y,z]
# return r: the rotation [x,y,z,w]
def read_tf_file(filename):
    if not os.path.exists(filename):
        print('File ' + filename + ' not found')
        return

    with open(filename, mode='r') as file:
        lines = file.readlines()

    t = [float(s) for s in lines[0].split()]
    r = [float(s) for s in lines[1].split()]

    return t, r


# reads lidar config and calibration data from a file
# param[in] calib_dir: base directory for dataset calibration files
# return params: dict with all lidar config and calibration parameters
def get_lidar_params(calib_dir):
    params = {'sensor_type': 'lidar', 'data_type': 'pointcloud'}
    filename = calib_dir + '/transforms/base_to_lidar.txt'
    params['translation'], params['rotation'] = read_tf_file(filename)
    return params


# reads imu config and calibration data from a file
# param[in] calib_dir: base directory for dataset calibration files
# return params: dict with all imu config and calibration parameters
def get_imu_params(calib_dir):
    params = {'sensor_type': 'imu'}
    filename = calib_dir + '/transforms/base_to_imu.txt'
    params['translation'], params['rotation'] = read_tf_file(filename)
    return params


# gets config data for groundtruth, doesn't read it from a file 
# because there's only two parameters and they're always the same
# return params: dict with the only two needed parameters for groundtruth
def get_groundtruth_params():
    return {'sensor_type': 'groundtruth', 'data_type': 'pose'}


# reads vicon config and calibration data from a file
# param[in] calib_dir: base directory for dataset calibration files
# return params: dict with all vicon config and calibration parameters
def get_vicon_params(calib_dir):
    params = {'sensor_type': 'vicon'}
    filename = calib_dir + '/transforms/base_to_vicon.txt'
    params['translation'], params['rotation'] = read_tf_file(filename)
    return params


# reads the antenna layout from the given file
# param[in] filename: the filename of the antenna config file
# return antenna_cfg: a dict containing antenna layout parameters
def read_antenna_cfg(filename):
    antenna_cfg = {}

    with open(filename, mode='r') as file:
        lines = file.readlines()

    for line in lines:
        vals = line.split()

        if vals[0] != '#':
            if vals[0] == 'num_rx':
                antenna_cfg['num_rx'] = int(vals[1])
                antenna_cfg['rx_locations'] = [0] * antenna_cfg['num_rx']
            elif vals[0] == 'num_tx':
                antenna_cfg['num_tx'] = int(vals[1])
                antenna_cfg['tx_locations'] = [0] * antenna_cfg['num_tx']
            elif vals[0] == 'rx':
                antenna_cfg['rx_locations'][int(vals[1])] = (int(vals[2]), int(vals[3]))
            elif vals[0] == 'tx':
                antenna_cfg['tx_locations'][int(vals[1])] = (int(vals[2]), int(vals[3]))
            elif vals[0] == 'F_design':
                antenna_cfg['F_design'] = float(vals[1])

    return antenna_cfg


# reads the heatmap config parameters from the given file and
# adds them to the given parameter dict
# param[in] hm_filename: the filename of the heatmap config file
# param[in] hm_params: dict containing the heatmap config
# return hm_params: heatmap config dict with heatmap config parameters added
def read_heatmap_cfg(hm_filename, hm_params):
    hm_params['data_type'] = 'heatmap'

    with open(hm_filename, mode='r') as file:
        lines = file.readlines()

    for line in lines:
        vals = line.split()

        if vals[0] == 'azimuth_bins' or vals[0] == 'elevation_bins':
            hm_params[vals[0]] = [float(s) for s in vals[1:]]
        elif vals[0] == 'range_bin_width':
            hm_params[vals[0]] = float(vals[1])
        else:
            hm_params[vals[0]] = int(vals[1])

    return hm_params


# reads the waveform config parameters from the given file and
# adds them to the given parameter dict
# param[in] wave_filename: the filename of the waveform config file
# param[in] wave_params: dict containing the waveform parameters
# return wave_params: waveform config dict with waveform config parameters added
def read_waveform_cfg(wave_filename, wave_params):
    wave_params['data_type'] = 'adc_samples'

    with open(wave_filename, mode='r') as file:
        lines = file.readlines()

    int_vals = ['num_rx', 'num_tx', 'num_adc_samples_per_chirp', 'num_chirps_per_frame']
    for line in lines:
        vals = line.split()
        if vals[0] in int_vals:
            wave_params[vals[0]] = int(vals[1])
        else:
            wave_params[vals[0]] = float(vals[1])

    return wave_params


# reads the antenna coupling calibration from the given file
# param[in] coupling_filename: the filename of the coupling config file
# return coupling_calib: dict containing antenna coupling calibration data
def read_coupling_cfg(coupling_filename):
    coupling_calib = {}

    with open(coupling_filename, mode='r') as file:
        lines = file.readlines()

    num_tx = int(lines[0].split(':')[1])
    num_rx = int(lines[1].split(':')[1])
    num_range_bins = int(lines[2].split(':')[1])
    coupling_calib['num_tx'] = num_tx
    coupling_calib['num_rx'] = num_rx
    coupling_calib['num_range_bins'] = num_range_bins
    coupling_calib['num_doppler_bins'] = int(lines[3].split(':')[1])
    data_str = lines[4].split(':')[1]
    data_arr = np.array(data_str.split(',')).astype('float')
    data_arr = data_arr[:-1:2] + 1j * data_arr[1::2]
    coupling_calib['data'] = data_arr.reshape(num_tx, num_rx, num_range_bins)

    return coupling_calib


# reads the phase and frequency calibration data from the given file
# param[in] calib_filename: the filename of the phase/frequency calibration file
# return phase_calib: dict containing the phase calibration data
# return freq_calib: dict containing the frequency calibration data
def read_phase_freq_calib(calib_filename):
    phase_calib = {}
    freq_calib = {}

    with open(calib_filename) as file:
        lines = file.readlines()

    num_tx = int(lines[0].split(':')[1])
    num_rx = int(lines[1].split(':')[1])
    frequency_slope = float(lines[2].split(':')[1])
    sampling_rate = float(lines[3].split(':')[1])

    phase_cal_mat_str = lines[5].split(':')[1]
    phase_cal_mat_arr = np.array(phase_cal_mat_str.split(',')).astype('float')
    phase_cal_mat_arr = phase_cal_mat_arr[:-1:2] + 1j * phase_cal_mat_arr[1::2]
    phase_cal_mat = phase_cal_mat_arr.reshape(num_tx, num_rx)

    phase_calib['num_rx'] = num_rx
    phase_calib['num_tx'] = num_tx
    phase_calib['cal_matrix'] = phase_cal_mat

    freq_cal_mat_str = lines[4].split(':')[1]
    freq_cal_mat_arr = np.array(freq_cal_mat_str.split(',')).astype('float')
    freq_cal_mat = freq_cal_mat_arr.reshape(num_tx, num_rx)

    freq_calib['num_rx'] = num_rx
    freq_calib['num_tx'] = num_tx
    freq_calib['frequency_slope'] = frequency_slope
    freq_calib['sampling_rate'] = sampling_rate
    freq_calib['cal_matrix'] = freq_cal_mat

    return phase_calib, freq_calib


# gets config and calibration data for the single chip radar sensor
# param[in] calib_dir: base directory for dataset calibration files
# return dict including waveform, heatmap, pointcloud, antenna, and coupling config data
def get_single_chip_params(calib_dir):
    wave_params = {'sensor_type': 'single_chip'}
    hm_params = {'sensor_type': 'single_chip'}
    pc_params = {'sensor_type': 'single_chip', 'data_type': 'pointcloud'}

    tf_filename = calib_dir + '//transforms//base_to_single_chip.txt'

    t, r = read_tf_file(tf_filename)

    wave_params['translation'] = t
    wave_params['rotation'] = r
    hm_params['translation'] = t
    hm_params['rotation'] = r
    pc_params['translation'] = t
    pc_params['rotation'] = r

    wave_filename = calib_dir + '/single_chip/waveform_cfg.txt'
    wave_params = read_waveform_cfg(wave_filename, wave_params)

    hm_filename = calib_dir + '/single_chip/heatmap_cfg.txt'
    hm_params = read_heatmap_cfg(hm_filename, hm_params)

    antenna_filename = calib_dir + '/single_chip/antenna_cfg.txt'
    antenna_cfg = read_antenna_cfg(antenna_filename)

    coupling_filename = calib_dir + '/single_chip/coupling_calib.txt'
    coupling_calib = read_coupling_cfg(coupling_filename)

    return {'waveform': wave_params,
            'heatmap': hm_params,
            'pointcloud': pc_params,
            'antenna': antenna_cfg,
            'coupling': coupling_calib}


# gets config and calibration data for the cascaded radar sensor
# param[in] calib_dir: base directory for dataset calibration files
# return dict including waveform, heatmap, antenna, coupling, 
#             phase, and frequency calibration data
def get_cascade_params(calib_dir):
    wave_params = {'sensor_type': 'cascade'}
    hm_params = {'sensor_type': 'cascade'}

    tf_filename = calib_dir + '/transforms/base_to_cascade.txt'

    t, r = read_tf_file(tf_filename)

    wave_params['translation'] = t
    wave_params['rotation'] = r
    hm_params['translation'] = t
    hm_params['rotation'] = r

    wave_filename = calib_dir + '/cascade/waveform_cfg.txt'
    wave_params = read_waveform_cfg(wave_filename, wave_params)

    hm_filename = calib_dir + '/cascade/heatmap_cfg.txt'
    hm_params = read_heatmap_cfg(hm_filename, hm_params)

    antenna_filename = calib_dir + '/cascade/antenna_cfg.txt'
    antenna_cfg = read_antenna_cfg(antenna_filename)

    # coupling_filename = calib_dir + '/cascade/coupling_calib.txt'
    # coupling_calib = read_coupling_cfg(coupling_filename)
    coupling_calib = {}  # placeholder since I haven't done a coupling calibration for the cascade sensor yet

    phase_filename = calib_dir + '/cascade/phase_frequency_calib.txt'
    phase_calib, freq_calib = read_phase_freq_calib(phase_filename)

    return {'waveform': wave_params,
            'heatmap': hm_params,
            'antenna': antenna_cfg,
            'coupling': coupling_calib,
            'phase': phase_calib,
            'frequency': freq_calib}



"""
Author:         Samuel Lovett
Date Created:   Apr 26, 2024
Last Edited:    Dec 06, 2024

Description:
Flexible script which performs Least-Squares and Temporally Weighted Least Squares regression depending on the arguments provided.
Program described executes the logic described in Sec. III of [1]

[1] Temporally Constrained Instanteneous Ego-Motion Estimation using 3D Doppler Radar
"""

import numpy as np


def build_A_matrix(theta_val, gamma_val):
    """
    Assemble the azimuth values into the A matrix in
    :param theta_val: theta values to build into an matrix
    :param gamma_val: gamma values to build into an matrix
    :return: A matrix used in the LSQ or TWLSQ solution to Ax=b
    """
    A_mat_col_0 = (np.cos(gamma_val) * np.cos(theta_val)).reshape(-1, 1)
    A_mat_col_1 = (np.cos(gamma_val) * np.sin(theta_val)).reshape(-1, 1)
    A_mat_col_3 = np.sin(gamma_val).reshape(-1, 1)
    A_mat = np.concatenate((A_mat_col_0, A_mat_col_1, A_mat_col_3), axis=1)
    return A_mat


def least_squares_velocity_analysis_3D(theta_val, gamma_val, doppler, weights=None):
    """
    Least-Squares and Temporally Weighted Least Squares regression depending on the arguments provided.

    :param theta_val: array of theta values
    :param gamma_val: array of theta values
    :param doppler: array of doppler values
    :param weights: array of weights corresponding to each theta value
    :return: Vx, Vy, Velocity, angle_rad
    """
    A = build_A_matrix(theta_val, gamma_val)
    b = doppler

    # calculate the weighted A and b matrix to perform W-LSQ
    if weights is not None:
        WA = weights[:, np.newaxis]
        A = A * np.sqrt(WA)
        b = b * np.sqrt(weights)

    Vx, Vy, Vz = np.linalg.lstsq(A, b, rcond=None)[0]

    return Vx, Vy, Vz



"""
Author:         Samuel Lovett
Date Created:   Apr 26, 2024
Last Edited:    Dec 06, 2024

Description:
Flexible script which performs squared error and Temporally Weighted Squared error calculations.
Program described executes (10) and (16) from [1]

[1] Temporally Constrained Instanteneous Ego-Motion Estimation using 3D Doppler Radar
"""
import numpy as np


def square_error(y_true, y_pred, weights=None):
    y_true = np.reshape(y_true, (-1, 1))
    y_pred = np.reshape(y_pred, (-1, 1))

    if weights is None:
        weight = np.ones(y_true.shape)
    else:
        weight = np.reshape(weights, (-1, 1))

    error_loss = weight * (y_pred - y_true) ** 2
    return error_loss


def mean_square_error(y_true, y_pred, weights=None):
    return np.sum(square_error(y_true, y_pred, weights)) / y_true.shape[0]


def l_one_norm(y_true, y_pred, weights=None):
    y_true = np.reshape(y_true, (-1, 1))
    y_pred = np.reshape(y_pred, (-1, 1))

    if weights is None:
        weight = np.ones(y_true.shape)
    else:
        weight = np.reshape(weights, (-1, 1))

    error_loss = weight * abs(y_pred - y_true)
    return error_loss


def mean_l_one_norm_error(y_true, y_pred, weights=None):
    return np.sum(l_one_norm(y_true, y_pred, weights)) / y_true.shape[0]



"""
Author:         Samuel Lovett
Date Created:   Apr 26, 2024
Last Edited:    Dec 06, 2024

Description:
Coloaradar Data importing script. Acts as an API between the coloradar dataset and the code in this repository

[1] Temporally Constrained Instanteneous Ego-Motion Estimation using 3D Doppler Radar
[2] Kramer A, Harlow K, Williams C, Heckman C. ColoRadar: The direct 3D millimeter wave radar dataset.
      The International Journal of Robotics Research. 2022;41(4):351-360. doi:10.1177/02783649211068535
"""

from tempEgo3D.dataset_loaders import *
import numpy as np
from scipy.spatial.transform import Rotation, Slerp
from tempEgo3D.radar_utilities import AIRE_filtering


class ColoradarDataset:
    """
    Coloaradar Data importing class. Acts as an API between the coloradar dataset and the code in this repository
    """
    def __init__(self, dataset_dir, calibration_directory='..//coloradar_package//dataset//calib', __n=3):
        self.dataset_dir = dataset_dir
        calibration_dir = calibration_directory

        all_radar_params = get_single_chip_params(calibration_dir)
        radar_params = all_radar_params['pointcloud']
        gt_params = get_groundtruth_params()

        # make transform from base frame to sensor frame as 4x4 transformation matrix
        H_bs = np.eye(4)
        H_bs[:3, 3] = radar_params['translation']
        H_bs[:3, :3] = Rotation.from_quat(radar_params['rotation']).as_matrix()
        self.H_bs = H_bs

        base_to_sensor_rot = np.array([[0, -1, 0],
                                       [1, 0, 0],
                                       [0, 0, 1]])
        self.base_to_sensor_rot_mat = base_to_sensor_rot

        radar_timestamps = get_timestamps(dataset_dir, radar_params)
        gt_timestamps = get_timestamps(dataset_dir, gt_params)
        gt_poses_unaligned = get_groundtruth(dataset_dir)

        self.gt_poses, self.radar_indices = interpolate_poses(gt_poses_unaligned, gt_timestamps, radar_timestamps)

        self.initial_time = radar_timestamps[self.radar_indices[0]]
        self.dataset_length = len(self.radar_indices)

        self.radar_params = radar_params
        self.radar_timestamps = radar_timestamps

        # set thresholds
        self.n = __n
        self.azimuth_thres_ = 80
        self.elevation_thres_ = 80
        self.range_thres_ = 0.3
        self.range_thres_max_ = 12
        self.intensity_thres_ = 1
        self.doppler_max_ = 2.56
        self.thresholds_ = np.array([self.azimuth_thres_, self.intensity_thres_,
                                     self.range_thres_, self.range_thres_max_,
                                     self.elevation_thres_, self.doppler_max_])

    def get_radar_cloud_3D(self, idx):
        """
        Method modified from COLORADAR package data importer [2]. \n
        Imports the radar cloud from the coloradar dataset and filters it using the AIRE filter.\n
        param[in] idx: measurement index to grab \n
        return radar_cloud: 5D list of filtered radar sample azimuth, elevation, doppler, range, and intensity values [filtered_theta, filtered_gamma, filtered_doppler, filtered_range, filtered_intensity] \n
        :param idx: measurement index to grab
        :return: radar_cloud
        """
        # Skips the first _ # of messages where there is no ground truth value and then uses the remaining messages
        # for hallways_run0 it skips the first 50
        matched_idx = self.radar_indices[idx]
        radar_pc_local = get_pointcloud(matched_idx, self.dataset_dir, self.radar_params)

        x = radar_pc_local[:, 0]
        y = radar_pc_local[:, 1]
        z = radar_pc_local[:, 2]
        intensity = radar_pc_local[:, 3]
        doppler_array = radar_pc_local[:, 4]
        range_val = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        theta = np.arctan2(y, x) # this is equivalent to atan(y/x) since x is always greater than 0
        xy = np.sqrt(x ** 2 + y ** 2)
        gamma = np.arctan2(z, xy)


        ## apply AIRE thresholding to remove near-field targets
        ## [azimuth, intensity, range, elevation, doppler]
        data_AIRE = np.column_stack((theta, intensity, range_val, gamma, doppler_array))

        idx_AIRE, Ntargets_valid = AIRE_filtering(data_AIRE, self.thresholds_)
        # print(Ntargets_valid)

        if Ntargets_valid < self.n:
            # print('Not enough valid points to estimate the velocity in TWLSQ')
            return None

        ## define pre-filtered radar data for further processing
        filtered_theta = theta[idx_AIRE]
        filtered_gamma = gamma[idx_AIRE]
        filtered_doppler = doppler_array[idx_AIRE]
        filtered_range = range_val[idx_AIRE]
        filtered_intensity = intensity[idx_AIRE]

        radar_cloud = [filtered_theta, filtered_gamma, filtered_doppler, filtered_range, filtered_intensity]

        return radar_cloud


    def get_gt_at(self, idx):
        """
        Get the ground truth translation and orientation for a given sample index.

        return translation: translation vector [x, y, z] in meters \n
        return orientation: 3x3 rotation matrix corresponding to the platforms orientation
        :param idx: sample index
        :return: translation, rotation
        """
        # GT location given as a homogeneous matrix with reference to the world frame
        orientation = self.gt_poses[idx][:3, :3]

        # Split into rotation and translation
        translation = self.gt_poses[idx][:3, 3]

        return translation, orientation

    def get_radar_time(self, idx):
        radar_timestamp_at_idx = self.radar_timestamps[self.radar_indices[idx]] - self.initial_time
        return radar_timestamp_at_idx

    def get_base_to_sensor_transformation_matrix(self):
        return self.base_to_sensor_rot_mat


def interpolate_poses(src_poses, src_stamps, tgt_stamps):
    """
    This code was written by the authors of [2] and is available at https://arpg.github.io/coloradar/ \n
    As described in [2]: Finds the ground truth pose of the platform at the radar timestamp.\n
    param[in] src_poses: list of poses in the form of a dict {'position': [x,y,z], 'orientation: [x,y,z,w]'} \n
    param[in] src_stamps: list of timestamps for the src_poses \n
    pram[in] tgt_stamps: list of timestamps for which poses need to be calculated \n
    return tgt_poses: list of interpolated poses as 4x4 transformation matrices \n
    return tgt_indices: list of indices in tgt_stamps for which poses were able to be interpolated \n
    :param src_poses: src_poses: list of poses in the form of a dict {'position': [x,y,z], 'orientation: [x,y,z,w]'}
    :param src_stamps: src_stamps: list of timestamps for the src_poses
    :param tgt_stamps: tgt_stamps: list of timestamps for which poses need to be calculated
    :return: tgt_poses, tgt_indices
    """
    src_start_idx = 0
    tgt_start_idx = 0
    src_end_idx = len(src_stamps) - 1
    tgt_end_idx = len(tgt_stamps) - 1

    # ensure first source timestamp is immediately before first target timestamp
    while tgt_start_idx < tgt_end_idx and tgt_stamps[tgt_start_idx] < src_stamps[src_start_idx]:
        tgt_start_idx += 1

    # ensure last source timestamp is immediately after last target timestamp
    while tgt_end_idx > tgt_start_idx and tgt_stamps[tgt_end_idx] > src_stamps[src_end_idx]:
        tgt_end_idx -= 1

    # iterate through target timestamps,
    # interpolating a pose for each as a 4x4 transformation matrix
    tgt_idx = tgt_start_idx
    src_idx = src_start_idx
    tgt_poses = []
    while tgt_idx <= tgt_end_idx and src_idx <= src_end_idx:
        # find source timestamps bracketing target timestamp
        while src_idx + 1 <= src_end_idx and src_stamps[src_idx + 1] < tgt_stamps[tgt_idx]:
            src_idx += 1

        # get interpolation coefficient
        c = ((tgt_stamps[tgt_idx] - src_stamps[src_idx])
             / (src_stamps[src_idx + 1] - src_stamps[src_idx]))

        # interpolate position
        pose = np.eye(4)
        pose[:3, 3] = ((1.0 - c) * src_poses[src_idx]['position']
                       + c * src_poses[src_idx + 1]['position'])

        # interpolate orientation
        r_src = Rotation.from_quat([src_poses[src_idx]['orientation'],
                                    src_poses[src_idx + 1]['orientation']])
        slerp = Slerp([0, 1], r_src)
        pose[:3, :3] = slerp([c])[0].as_matrix()

        tgt_poses.append(pose)

        # advance target index
        tgt_idx += 1

    tgt_indices = range(tgt_start_idx, tgt_end_idx + 1)

    return tgt_poses, tgt_indices


"""
This code has been developed by
Kramer A, Harlow K, Williams C, Heckman C. ColoRadar: The direct 3D millimeter wave radar dataset.
The International Journal of Robotics Research. 2022;41(4):351-360. doi:10.1177/02783649211068535

Available at https://arpg.github.io/coloradar/
"""

import numpy as np
from scipy.fft import fft
import math

# returns sample i for a blackman window of size n
# param[in] i: sample index
# param[in] n: size of the window
# return: value of sample i of a blackman window of size n
def blackman(i, n):
  a0 = 0.42
  a1 = 0.5
  a2 = 0.08
  return (a0 - a1 * math.cos((2.0 * math.pi * i) / n)
    + a2 * math.cos((4.0 * math.pi * i) / n))


# applies antenna coupling calibration
# param[in] adc_samples: raw complex adc samples for a frame of radar data
# param[in] coupling_calib: antenna coupling calibration returned 
#                           from dataset_loaders.get_cascade_params()\
# return: post-range-fft radar data with antenna coupling calibration applied
def apply_coupling_calibration(adc_samples, coupling_calib):

  num_tx = coupling_calib['num_tx']
  num_rx = coupling_calib['num_rx']
  num_adc_samples_per_chirp = adc_samples.shape[3]
  num_range_bins = coupling_calib['num_range_bins']
  num_doppler_bins = coupling_calib['num_doppler_bins']

  if num_tx != adc_samples.shape[0]:
    print('provided calibration data dimension 0 does not match data frame dimension 0')
    return None
  if num_rx != adc_samples.shape[1]:
    print('provided calibration data dimension 1 does not match data frame dimension 1')
    return None
  if num_range_bins != num_adc_samples_per_chirp // 2:
    print('provided calibration data dimension 2 does not match data frame dimension 3')
    return None

  # calculate range window
  range_window =  np.zeros(num_adc_samples_per_chirp)
  for i in range(num_adc_samples_per_chirp):
    range_window[i] = blackman(np.double(i), np.double(num_adc_samples_per_chirp))

  # apply range window
  adc_samples *= range_window.reshape(1,1,1,-1)

  # run range fft
  range_fft = fft(adc_data, n=num_adc_samples_per_chirp,axis=-1)

  # discard negative component of range fft output
  range_fft = range_fft[:,:,:,1:num_range_bins]

  # apply coupling calibration
  return range_fft - coupling_calib['data'].reshape(num_tx, num_rx, 1, num_range_bins)


# applies phase calibration 
# param[in] adc_samples: raw complex adc samples for a frame of radar data
# param[in] phase_calib: phase calibration returned from dataset_loaders.get_cascade_params()
# return: adc samples with phase calibration applied
def apply_phase_calibration(adc_samples, phase_calib):

  num_tx = phase_calib['num_tx']
  num_rx = phase_calib['num_rx']
  phase_cal_mat = phase_calib['cal_matrix']

  # create phase calibration matrix
  phase_cal_mat = phase_cal_mat[0,0] / phase_cal_mat

  # apply phase calibration
  phase_calibrated_adc_samples = adc_samples * phase_cal_mat.reshape(num_tx, num_rx, 1, 1)

  return phase_calibrated_adc_samples


# applies frequency calibration
# param[in] adc_samples: raw complex adc samples for a frame of radar data
# param[in] freq_calib: frequency calibration returned from dataset_loaders.get_cascade_params()
# param[in] wave_config: waveform config returned from dataset_loaders.get_cascade_params()
# return: adc samples with frequency calibration applied
def apply_frequency_calibration(adc_samples, freq_calib, wave_config):

  num_tx = freq_calib['num_tx']
  num_rx = freq_calib['num_rx']
  cal_frequency_slope = freq_calib['frequency_slope']
  cal_sampling_rate = freq_calib['sampling_rate']
  frequency_slope = wave_config['frequency_slope']
  sampling_rate = wave_config['adc_sample_frequency']
  num_adc_samples_per_chirp = wave_config['num_adc_samples_per_chirp']
  freq_cal_mat = freq_calib['cal_matrix']

  # create frequency calibration matrix
  delta_p = freq_cal_mat - freq_cal_mat[0,0]
  freq_calib = (2.0 * math.pi * delta_p 
                * (frequency_slope / cal_frequency_slope)
                * (cal_sampling_rate / sampling_rate))
  freq_calib /= num_adc_samples_per_chirp
  freq_calib_vec = np.expand_dims(freq_calib, -1) * np.arange(num_adc_samples_per_chirp)
  freq_calib_vec = np.exp(-1j * freq_calib_vec)

  # apply frequency calibration
  freq_calibrated_adc_samples = freq_calib_vec.reshape(num_tx,num_rx,1,num_adc_samples_per_chirp) * adc_samples

  return freq_calibrated_adc_samples



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name='tempEgo3D',
    version='0.1.0',
    packages=find_packages(),
    install_requires=[
        'numpy>=1.23.1',
        'scipy>=1.10.1',

    ],
    entry_points={
        'console_scripts': [],
    },
    url='https://github.com/samuelLovett/tempEgo3D.git',
    author='Samuel Lovett',
    author_email='samuellovett@cmail.carleton.ca',
    description="A package for estimating 3D ego-motion velocity using millimetre-wave (mmWave) radar",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)


"""
Author:         Samuel Lovett
Date Created:   Apr 26, 2024
Last Edited:    Dec 06, 2024

Description:
Estimate the 3D body-frame velocity vector of the sensor platform given input data from a single radar as described in
[1]. For use with your own radar sensor modify the code in the update_buffer_weighted method.

The velocity estimation scheme takes the following approach:

1. A weighting is applied to all datapoints within the sliding window based on the datapoints corresponding
measurement-index

2. A RANSAC outlier rejection method is used to filter the data which contains a large amount of
outliers. RANSAC selects points uniformly or based on their weights to estimate the velocity. This process is then
repeated until a first-pass velocity estimate is derived from the best inlier set.

3. Least Squares Distance Regression (LSQ) is initialized with the inlier set and used to calculate the final velocity
estimate of the body frame linear velocity.

[1] Temporally Constrained Instanteneous Ego-Motion Estimation using 3D Doppler Radar
"""
from tempEgo3D.error_and_loss_function import *
from tempEgo3D.external_dataset_import import *
from tempEgo3D.ego_motion_methods import *
import numpy as np
from collections import deque
import time


def set_RANSAC(dataset):
    """
    Define the RANSAC method parameter's as described in Sec. III Hyperparameter optimization of [1]
    :param dataset_key: key to use with the dataset directory dictionary
    :return: The velocity estimation object
    """
    n = 3  # Minimum number of data points to estimate parameters
    k = 608  # Maximum iterations allowed
    epsilon = 0.0006194819604855642  # Threshold value to determine if points are fit well
    rho = 31  # Number of points to assert that the model fits well
    start_idx = 0
    stop_idx = None
    method = VelocityEstimation(dataset, start_idx, stop_idx, n, k, epsilon, rho)
    return method


# def set_MLESAC_buffered(dataset):
#     """
#     Define the MLESAC method parameter's as described in
#     :return: The velocity estimation object
#     """
#     m = 2  # Buffer size
#     n = 3  # Minimum number of data points to estimate parameters
#     start_idx = 0
#     stop_idx = None
#     method = MLESACBufferVelEstimate(dataset, start_idx, stop_idx, n, m)
#     return method


def set_TWLSQ(dataset):
    """
    Define the Temporally Weighted Least Squares method parameter's as described in Sec. III Hyperparameter
    optimization of [1]
    :param dataset_key: key to use with the dataset directory dictionary
    :return: The velocity estimation object
    """
    m = 2  # Buffer size
    fff_lambda = 0.15760037633371646  # 0 values only the most recent samples and 1 values all samples equally
    n = 3  # Minimum number of data points to estimate parameters
    k = 68  # Maximum iterations allowed
    epsilon = 14.39032148114001  # Threshold value to determine if points are fit well
    rho = 24  # Number of points to assert that the model fits well
    start_idx = 0
    stop_idx = None

    method = TWLSQVelEstimate(dataset,
                              start_idx,
                              stop_idx,
                              m,
                              fff_lambda,
                              n,
                              k,
                              epsilon,
                              rho)
    return method


def set_TEMPSAC(dataset):
    """
    Define the TEMPoral SAmpling Concensus parameter's as described in Sec. III Hyperparameter
    optimization of [1]
    :param dataset_key: key to use with the dataset directory dictionary
    :return: The velocity estimation object
    """
    m = 2  # Buffer size
    fff_lambda = 0.7160085793075731  # 0 values only the most recent samples and 1 values all samples equally
    n = 3  # Minimum number of data points to estimate parameters
    k = 1494  # Maximum iterations allowed
    epsilon = 14.843480780976819  # Threshold value to determine if points are fit well
    rho = 24  # Number of points to assert that the model fits well
    start_idx = 0
    stop_idx = None

    method = TEMPSACVelEstimate(dataset,
                                start_idx,
                                stop_idx,
                                m,
                                fff_lambda,
                                n,
                                k,
                                epsilon,
                                rho)

    return method


def get_subset_of_weights_normalized(buffer_length, fff_lambda):
    """
    calculates the normalized weight value for each measurement index when the sliding window is not full
    :param buffer_length: length of the sliding window labelled as m in [1]
    :param fff_lambda: fixed factor lambda as defined in (14) of [1]
    :return: list of weight values normalized by the number of measurements in the sliding window
    """

    weights = []
    buffer_length_indexed_from_zero = buffer_length - 1
    for buffer_index in range(buffer_length):
        weight_at_i = fff_lambda ** (buffer_length_indexed_from_zero - buffer_index)
        weights.append(weight_at_i)

    sum_of_weights = sum(weights)
    normalized_weights = [weights[i] / sum_of_weights for i in range(len(weights))]
    return normalized_weights


def get_subset_of_weights(buffer_length, fff_lambda):
    """
    calculates the normalized weight value for each measurement index when the sliding window is not full
    :param buffer_length: length of the sliding window labelled as m in [1]
    :param fff_lambda: fixed factor lambda as defined in (14) of [1]
    :return: list of weight values
    """

    weights = []
    buffer_length_indexed_from_zero = buffer_length - 1
    for buffer_index in range(buffer_length):
        weight_at_i = fff_lambda ** (buffer_length_indexed_from_zero - buffer_index)
        weights.append(weight_at_i)

    return weights


class VelocityEstimation:
    """
    Standard velocity estimation class which implements the RANSAC method as shown in [1]
    """

    def __init__(self, __dataset, __start_idx, __stop_idx, __n, __k, __epsilon, __rho):
        """
        Initialization of the standard velocity estimation method RANSAC as described at the beginning of this code.
        :param __dataset: the dataset directory dictionary
        :param __start_idx: measurement index to start the evaluation at
        :param __stop_idx: measurement index to stop the evaluation at
        :param __n: Minimum number of samples to estimate parameters
        :param __k: Maximum iterations allowed
        :param __epsilon: Threshold value to determine if points are fit well
        :param __rho: Number of close samples required to assert model fits well
        """
        calib_dir = 'coloradar_package//dataset//calib'
        self.coloradar_dataset = ColoradarDataset(__dataset, calib_dir)
        if __stop_idx is None:
            self.stop = self.coloradar_dataset.dataset_length
        else:
            self.stop = __stop_idx
        self.start_idx = __start_idx

        self.velocity_estimator = RANSAC(loss=square_error,
                                                metric=mean_square_error,
                                                n=__n,
                                                k=__k,
                                                epsilon=__epsilon,
                                                rho=__rho)

    def update_buffer(self, idx):
        """
        Update the sliding window as described in Sec. III B. of [1]. Index m is the most recent measurement in time and index 0 is the oldest measurement in time. \n
        \n param[in] idx: measurement index to import from coloradar
        \n return tensor_stack: multidimensional array dimension 1 is [theta, gamma, vd], and dimension 2 is the sample index
        \n return weight_mask: a list of weight values corresponding to each sample in tensor_stack.\n
        For use with your own data replace the line: current_sample = self.coloradar_dataset.get_radar_cloud(idx) and data must be formatted as a 3D list [theta, gamma, doppler].
        \t theta is a list of azimuth values in rads. \n
        \t gamma is a list of elevation values in rads. \n
        \t doppler is a list of doppler velocity values in m/s.

        :param idx: current measurement index used to import values from the dataset
        :return: current_sample
        """

        # Get the new measurement
        current_sample = self.coloradar_dataset.get_radar_cloud_3D(idx)
        return current_sample


# class MLESACBufferVelEstimate(VelocityEstimation):
#     """
#     MLESAC Buffered velocity estimation class which inherits the initialization of the standard method but uses MLESAC from
#     goggles.
#     """
#
#     def __init__(self, __dataset, __start_idx, __stop_idx, __buffer_size, __n):
#         super().__init__(__dataset, __start_idx, __stop_idx, __n, None, None, None)
#         if __dataset_key is not None:
#             __results_location = 'results//buffer//' + __dataset_key + '//mlesacBuffered//'
#             self.my_results = MLESACBufferedResults(__results_location)
#         self.my_opt_results = OptResultsClass()
#
#         self.velocity_estimator = MLESACEstimator()
#         self.buffer_size = __buffer_size
#         self.sliding_window = []
#
#     def update_buffer(self, idx):
#         """
#         Update the sliding window as described in Sec. II C. of [1]. Index m is the most recent measurement in time and index 0 is the oldest measurement in time. \n
#         \n param[in] idx: measurement index to import from coloradar
#         \n return current measurement: multidimensional array dimension 1 is [theta, vd], and dimension 2 is the sample index
#         For use with your own data replace the line: current_sample = self.coloradar_dataset.get_radar_cloud(idx) and data must be formatted as a 2D list [theta, doppler].
#         \t theta is a list of azimuth values in rads. \n
#         \t doppler is a list of doppler velocity values in m/s.
#
#         :param idx: current measurement index used to import values from the dataset
#         :return: current measurement
#         """
#         # Get the new measurement
#         current_sample = self.coloradar_dataset.get_radar_cloud_3D(idx)
#         buffer_length = len(self.sliding_window)
#
#         current_buffer = self.sliding_window
#
#         # append the current measurement to the buffer
#         if buffer_length < self.buffer_size:
#             current_buffer.append(current_sample)
#             buffer_length = buffer_length + 1
#         else:
#             # Remove the oldest measurement from the list
#             # Update the buffer to include the new measurement
#             current_buffer.pop(0)
#             current_buffer.append(current_sample)
#
#         # Assemble the multidimensional array
#         reshaped_tensor_stack = np.concatenate(list(current_buffer), axis=1)
#         theta = reshaped_tensor_stack[0, :]
#         gamma = reshaped_tensor_stack[1, :]
#         doppler = reshaped_tensor_stack[2, :]
#         object_range = reshaped_tensor_stack[3, :]
#         intensity = reshaped_tensor_stack[4, :]
#         return theta, gamma, doppler, object_range, intensity


class TWLSQVelEstimate(VelocityEstimation):
    """
    TWLSQ velocity estimation class which inherits the initialization of the standard method but uses TWLSQ_RANSAC
    as shown in [1].
    """

    def __init__(self, __dataset, __start_idx, __stop_idx, __buffer_size,
                 __fff_lambda, __n, __k, __epsilon, __rho):
        """
        Initialization of the velocity estimation method as described at the beginning of this code.
        :param __dataset_key: key to use with the dataset directory dictionary
        :param __start_idx: measurement index to start the evaluation at
        :param __stop_idx: measurement index to stop the evaluation at
        :param __buffer_size: Size of the sliding window, m in [1]
        :param __fff_lambda: fixed forgetting factor as defined in (14) of [1]. 0 uses only the most recent samples and 1 uses all samples equally.
        :param __n: Minimum number of samples to estimate parameters
        :param __k: Maximum iterations allowed
        :param __epsilon: Threshold value to determine if points are fit well
        :param __rho: Number of close samples required to assert model fits well
        """
        super().__init__(__dataset, __start_idx, __stop_idx, __n, __k, __epsilon, __rho)

        self.velocity_estimator = RANSAC_TWLSQ(loss=square_error,
                                                      metric=mean_square_error,
                                                      n=__n,
                                                      k=__k,
                                                      epsilon=__epsilon,
                                                      rho=__rho)

        fff_lambda = __fff_lambda

        weights = []
        buffer_size_indexed_from_zero = __buffer_size - 1
        for buffer_index in range(__buffer_size):
            weight_at_i = fff_lambda ** (buffer_size_indexed_from_zero - buffer_index)
            weights.append(weight_at_i)

        self.exponential_weights_list = weights

        self.buffer_size = __buffer_size
        self.sliding_window = []
        self.fff_lambda = fff_lambda

    def update_buffer(self, idx):
        """
        Update the sliding window as described in Sec. III B. of [1]. Index m is the most recent measurement in time and index 0 is the oldest measurement in time. \n
        \n param[in] idx: measurement index to import from coloradar
        \n return tensor_stack: multidimensional array dimension 1 is [theta, gamma, vd], and dimension 2 is the sample index
        \n return weight_mask: a list of weight values corresponding to each sample in tensor_stack.\n
        For use with your own data replace the line: current_sample = self.coloradar_dataset.get_radar_cloud(idx) and data must be formatted as a 2D list [theta, doppler].
        \t theta is a list of azimuth values in rads. \n
        \t gamma is a list of elevation values in rads. \n
        \t doppler is a list of doppler velocity values in m/s.

        :param idx: current measurement index used to import values from the dataset
        :return: tensor_stack, weight_mask
        """

        # Get the new measurement
        current_sample = self.coloradar_dataset.get_radar_cloud_3D(idx)

        if current_sample is None:
            return None

        buffer_length = len(self.sliding_window)

        current_buffer = self.sliding_window

        # append the current measurement to the buffer
        if buffer_length < self.buffer_size:
            current_buffer.append(current_sample)
            buffer_length = buffer_length + 1
            # weights = get_subset_of_weights_normalized(buffer_length, self.fff_lambda)
            weights = get_subset_of_weights(buffer_length, self.fff_lambda)
        else:
            # Remove the oldest measurement from the list
            # Update the buffer to include the new measurement
            current_buffer.pop(0)
            current_buffer.append(current_sample)
            # weights = self.normalized_exponential_weights_list
            weights = self.exponential_weights_list

        # update the corresponding weight for each sample within the buffer depending on its measurement index
        weight_mask = np.array([])
        for i in range(buffer_length):
            mask_at_i = np.ones(len(current_buffer[i][0])) * weights[i]
            weight_mask = np.concatenate((weight_mask, mask_at_i))

        # Assemble the multidimensional array
        reshaped_tensor_stack = np.concatenate(list(current_buffer), axis=1)
        return reshaped_tensor_stack[0, :], reshaped_tensor_stack[1, :], reshaped_tensor_stack[2,
                                                                         :], weight_mask  # break it into theta, vd, and w


class TEMPSACVelEstimate(VelocityEstimation):
    """
    TEMPSAC velocity estimation class which inherits the initialization of the standard method and the TWLSQ method but
    uses TEMPSAC as shown in [1].
    """

    def __init__(self, __dataset, __start_idx, __stop_idx, __buffer_size,
                 __fff_lambda, __n, __k, __epsilon, __rho):
        """
        Initialization of the velocity estimation method as described at the beginning of this code.
        :param __dataset_key: key to use with the dataset directory dictionary
        :param __start_idx: measurement index to start the evaluation at
        :param __stop_idx: measurement index to stop the evaluation at
        :param __buffer_size: Size of the sliding window, m in [1]
        :param __fff_lambda: fixed forgetting factor as defined in (14) of [1]. 0 uses only the most recent samples and 1 uses all samples equally.
        :param __n: Minimum number of samples to estimate parameters
        :param __k: Maximum iterations allowed
        :param __epsilon: Threshold value to determine if points are fit well
        :param __rho: Number of close samples required to assert model fits well
        """
        super().__init__(__dataset,
                         __start_idx,
                         __stop_idx,
                         __n,
                         __k,
                         __epsilon,
                         __rho)
        self.velocity_estimator = TEMPSAC(loss=square_error,
                                                 metric=mean_square_error,
                                                 n=__n,
                                                 k=__k,
                                                 epsilon=__epsilon,
                                                 rho=__rho)

        fff_lambda = __fff_lambda


        weights = []
        buffer_size_indexed_from_zero = __buffer_size - 1
        for buffer_index in range(__buffer_size):
            weight_at_i = fff_lambda ** (buffer_size_indexed_from_zero - buffer_index)
            weights.append(weight_at_i)
        sum_of_weights = sum(weights)
        normalized_weights = [weights[i] / sum_of_weights for i in range(len(weights))]

        self.normalized_exponential_weights_list = normalized_weights

        self.buffer_size = __buffer_size
        self.sliding_window = []
        self.fff_lambda = fff_lambda

    def update_buffer(self, idx):
        """
        Update the sliding window as described in Sec. III B. of [1]. Index m is the most recent measurement in time and index 0 is the oldest measurement in time. \n
        \n param[in] idx: measurement index to import from coloradar
        \n return tensor_stack: multidimensional array dimension 1 is [theta, gamma, vd], and dimension 2 is the sample index
        \n return weight_mask: a list of weight values corresponding to each sample in tensor_stack.\n
        For use with your own data replace the line: current_sample = self.coloradar_dataset.get_radar_cloud(idx) and data must be formatted as a 2D list [theta, doppler].
        \t theta is a list of azimuth values in rads. \n
        \t gamma is a list of elevation values in rads. \n
        \t doppler is a list of doppler velocity values in m/s.

        :param idx: current measurement index used to import values from the dataset
        :return: tensor_stack, weights
        """

        current_sample = self.coloradar_dataset.get_radar_cloud_3D(idx)  # Get the new item from the sensor

        buffer_length = len(self.sliding_window)
        current_buffer = self.sliding_window

        if buffer_length < self.buffer_size:
            current_buffer.append(current_sample)
            buffer_length = buffer_length + 1
            weights = get_subset_of_weights_normalized(buffer_length, self.fff_lambda)
        else:
            current_buffer.pop(0)  # Remove the oldest item from the list
            current_buffer.append(current_sample)  # Update the buffer to include the new item
            weights = self.normalized_exponential_weights_list

        weight_mask = np.array([])
        for i in range(buffer_length):
            mask_at_i = np.ones(len(current_buffer[i][0])) * weights[i]
            weight_mask = np.concatenate((weight_mask, mask_at_i))

        return current_buffer, weights



def main():
    """
    Example main function. \n
    Modify the paths contained in datasets_dic and calib_dir defined in the class VelocityEstimation __init__ function if using a different directory structure than described in the readme.
    """

    datasets_dic = {'army_run0': 'coloradar_package/dataset/2_23_2021_edgar_army_run0/',
                    'army_run1': 'coloradar_package/dataset/2_23_2021_edgar_army_run1/',
                    'army_run2': 'coloradar_package/dataset/2_23_2021_edgar_army_run2/',
                    'army_run3': 'coloradar_package/dataset/2_23_2021_edgar_army_run3/',
                    'army_run4': 'coloradar_package/dataset/2_23_2021_edgar_army_run4/',
                    'classroom_run2': 'coloradar_package/dataset/2_23_2021_edgar_classroom_run2/',
                    'classroom_run3': 'coloradar_package/dataset/2_23_2021_edgar_classroom_run3/',
                    'classroom_run4': 'coloradar_package/dataset/2_23_2021_edgar_classroom_run4/',
                    'classroom_run5': 'coloradar_package/dataset/2_23_2021_edgar_classroom_run5/',
                    'hallways_run0': 'coloradar_package/dataset/12_21_2020_ec_hallways_run0/',
                    'hallways_run1': 'coloradar_package/dataset/12_21_2020_ec_hallways_run1/',
                    'hallways_run2': 'coloradar_package/dataset/12_21_2020_ec_hallways_run2/',
                    'hallways_run3': 'coloradar_package/dataset/12_21_2020_ec_hallways_run3/',
                    'irl_run0': 'coloradar_package/dataset/12_21_2020_irl_lab_run0/',
                    'irl_run1': 'coloradar_package/dataset/12_21_2020_irl_lab_run1/',
                    'irl_run3': 'coloradar_package/dataset/12_21_2020_irl_lab_run3/',
                    'irl_run4': 'coloradar_package/dataset/12_21_2020_irl_lab_run4/'
                    }

    key = 'classroom_run2'
    dataset = datasets_dic[key]

    start_time = time.time()

    standard = set_RANSAC(dataset)
    TWLSQ = set_TWLSQ(dataset)
    TEMPSAC = set_TEMPSAC(dataset)

    """
    Enter your code here. The following loop calculates the ego velocity of the platform.
    """
    ransac = standard.velocity_estimator
    idx = standard.start_idx
    stop = standard.stop

    while idx < stop:
        radar_cloud = standard.update_buffer(idx)
        velocities_at_idx = ransac.estimate_velocity(radar_cloud)
        print(velocities_at_idx)
        idx = idx + 1

    end_time = time.time()
    elapsed_time = end_time - start_time

    print(f"Elapsed time: {elapsed_time} seconds")


if __name__ == '__main__':
    main()



"""
Author:         Samuel Lovett
Date Created:   Apr 26, 2024
Last Edited:    Dec 06, 2024

Description:
Assemble the multidimensional array and break it into data axis corresponds to [theta, gamma, velocity]. This will seem odd in
the context of the code because it is. I assemble the tensor stack just to dismantle it. But it was the easiest way to
combine all of the measurements into one long vector for each data axis while still
preserving the shape of the buffer needed for TEMPSAC initial selection.

[1] Temporally Constrained Instanteneous Ego-Motion Estimation using 3D Doppler Radar

"""

import numpy as np


def data_unpacker_tensor_object_3D(structured_data):
    """
    param[in] structured_data: Sliding window as a list of measurements

    \n return azimuth_val: List of azimuth values corresponding to each sample in the sliding window
    \n return elevation_val: List of elevation values corresponding to each sample in the sliding window
    \n return doppler_val: List of doppler values corresponding to each sample in the sliding window
    :param structured_data: Sliding window as a list of measurements
    :return: azimuth_val, elevation_val, doppler_val
    """
    tensor_stack = np.concatenate(list(structured_data), axis=1)
    azimuth_val = tensor_stack[0, :]
    elevation_val = tensor_stack[1, :]
    doppler_val = tensor_stack[2, :]

    if doppler_val.size == azimuth_val.size:
        # print(doppler_val.size)
        return azimuth_val, elevation_val, doppler_val
    else:
        raise ValueError("The size of the doppler_val array does not match the theta array")